from calendar import monthrange
from dataclasses import dataclass, field
from datetime import date
from decimal import Decimal
from enum import Enum
from typing import List, Optional, Dict
from pydantic import BaseModel


class TransactionOperation(Enum):
    CREDIT = 'credit'
    DEBIT = 'debit'
    SET = 'set'


class DataType(Enum):
    NUMBER = 'number'
    STRING = 'string'
    BOOLEAN = 'boolean'


class ScheduleEndType(Enum):
    NO_END = "no_end"
    END_REPEATS = "end_repeats"
    END_DATE = "end_date"


class BusinessDayAdjustment(Enum):
    NO_ADJUSTMENT = "no_adjustment"
    NEXT_WORKING_DAY = "next_working_day"
    PREVIOUS_WORKING_DAY = "previous_working_day"
    CLOSEST_WORKING_DAY = "closest_working_day"


class ScheduleFrequency(Enum):
    DAILY = "daily"
    MONTHLY = "monthly"


class ScheduledTransactionTiming(Enum):
    START_OF_DAY = "start_of_day"
    END_OF_DAY = "end_of_day"


class PositionType(BaseModel):
    name: str
    label: str


class RateTier(BaseModel):
    from_amount: Decimal
    to_amount: Decimal
    rate: Decimal


class RateType(BaseModel):
    name: str
    label: str
    rate_tiers: List[RateTier] = []

    def add_tier(self, to_amount: Decimal, rate: Decimal):
        rate_tier = RateTier(from_amount=self.get_max_to_amount(), to_amount=to_amount, rate=rate)
        self.rate_tiers.append(rate_tier)

    def get_max_to_amount(self):
        if len(self.rate_tiers) == 0:
            return Decimal(0)

        max_value = max(self.rate_tiers, key=lambda tier: tier.to_amount)
        return max_value.to_amount

    def get_rate(self, amount: Decimal):
        rate_tier = next(rt for rt in self.rate_tiers if rt.from_amount <= amount <= rt.to_amount)

        return rate_tier.rate

    def get_daily_fee(self, users: Decimal, value_date: date):
        _, days_in_month = monthrange(value_date.year, value_date.month)
        monthly_fee = Decimal(self.get_fee(Decimal(0), users))
        return monthly_fee / Decimal(days_in_month)

    def get_fee(self, from_amount: Decimal, to_amount: Decimal):
        processed = from_amount
        fee = Decimal(0)

        exit_loop = False

        for rate_tier in self.rate_tiers:
            if rate_tier.from_amount <= processed < rate_tier.to_amount:
                part_processed = rate_tier.to_amount - processed

                if to_amount < rate_tier.to_amount:
                    part_processed = to_amount - processed
                    exit_loop = True

                fee = fee + part_processed * rate_tier.rate
                processed = processed + part_processed

                if exit_loop:
                    break

        return fee


class PositionRule(BaseModel):
    operation: TransactionOperation
    position_type_name: str


class TransactionType(BaseModel):
    name: str
    label: str
    position_rules: List[PositionRule] = []

    def add_position_rule(self, transaction_operation: TransactionOperation, position_type: PositionType):
        self.position_rules.append(PositionRule(operation=transaction_operation, position_type_name=position_type.name))
        return self


class ScheduleType(BaseModel):
    name: str
    label: str
    frequency: ScheduleFrequency
    end_type: ScheduleEndType
    business_day_adjustment: BusinessDayAdjustment
    interval_expression: str
    start_date_expression: str
    end_date_expression: str = None
    number_of_repeats_expression: str = None
    include_dates_expression: str = None
    exclude_dates_expression: str = None


class ScheduledTransaction(BaseModel):
    schedule_name: str
    timing: ScheduledTransactionTiming
    generated_transaction_type: str
    amount_expression: str


class TriggeredTransaction(BaseModel):
    trigger_transaction_type_name: str
    generated_transaction_type: str
    amount_expression: str


class PropertyType(BaseModel):
    name: str
    label: str


class AccountType(BaseModel):
    name: str
    label: str
    transaction_type_names: List[str] = []
    schedule_types: List[ScheduleType] = []
    property_types: List[PropertyType] = []
    scheduled_transactions: List[ScheduledTransaction] = []
    triggered_transactions: List[TriggeredTransaction] = []

    def add_transaction_type(self, transaction_type: TransactionType):
        self.transaction_type_names.append(transaction_type.name)

    def add_schedule_type(self, schedule_type: ScheduleType):
        self.schedule_types.append(schedule_type)

    def add_scheduled_transaction(self, schedule_type: ScheduleType, timing: ScheduledTransactionTiming,
                                  generated_transaction_type: TransactionType, amount_expression: str):
        scheduled_transaction = ScheduledTransaction(schedule_name=schedule_type.name, timing=timing,
                                                     generated_transaction_type=generated_transaction_type.name,
                                                     amount_expression=amount_expression)
        self.scheduled_transactions.append(scheduled_transaction)

    def add_trigger_transaction(self, trigger_transaction_type: TransactionType,
                                generated_transaction_type: TransactionType, amount_expression: str):
        triggered_transaction = TriggeredTransaction(name=trigger_transaction_type.name,
                                                     generated_transaction_type=generated_transaction_type.name,
                                                     amount_expression=amount_expression)
        self.triggered_transactions.append(triggered_transaction)


class Configuration(BaseModel):
    version: str
    transaction_types: List[TransactionType] = []
    position_types: List[PositionType] = []
    account_types: List[AccountType] = []
    rate_types: Dict[str, RateType] = {}
    triggered_transactions: List[TriggeredTransaction] = []

    def add_transaction_type(self, name: str, label: str) -> TransactionType:
        transaction_type = TransactionType(name=name, label=label)
        self.transaction_types.append(transaction_type)
        return transaction_type

    def add_position_type(self, name: str, label: str) -> PositionType:
        position_type = PositionType(name=name, label=label)
        self.position_types.append(position_type)
        return position_type

    def add_account_type(self, name: str, label: str) -> AccountType:
        account_type = AccountType(name=name, label=label)
        self.account_types.append(account_type)
        return account_type

    def add_rate_type(self, name: str, label: str) -> RateType:
        rate_type = RateType(name=name, label=label)
        self.rate_types[name] = rate_type
        return rate_type

    def add_trigger_transaction(self, trigger_transaction_type: TransactionType,
                                generated_transaction_type: TransactionType, amount_expression: str):
        trigger_transaction = TriggeredTransaction(trigger_transaction_type_name=trigger_transaction_type.name,
                                                   generated_transaction_type=generated_transaction_type.name,
                                                   amount_expression=amount_expression)
        self.triggered_transactions.append(trigger_transaction)

    def get_transaction_type(self, transaction_type_name: str) -> TransactionType:
        return next(tt for tt in self.transaction_types if tt.name == transaction_type_name)

    def get_account_type(self, account_type_name: str):
        return next(at for at in self.account_types if at.name == account_type_name)

    def get_rate_type(self, rate_type_name: str):
        return next(rt for rt in self.rate_types if rt.name == rate_type_name)

    def get_trigger_transaction(self, trigger_transaction_type_name: str) -> Optional[TriggeredTransaction]:
        return next((tt for tt in self.triggered_transactions if
                     tt.trigger_transaction_type_name == trigger_transaction_type_name), None)

    def __getattr__(self, method_name: str):
        # find rate type by method name

        if method_name in self.rate_types:
            return self.rate_types[method_name]
        else:
            raise AttributeError(f'No such attribute: {method_name}')



import unittest
from datetime import date
from decimal import Decimal

from accounts.metadata import Configuration
from accounts.runtime import *
from tests.test_config import create_savings_account


class TestAccount(unittest.TestCase):
    def test_create_account(self):
        config = create_savings_account()

        account_type = config.account_types[0]

        account = Account(start_date=date.today(), account_type_name=account_type.name, config_version=config.version,
                          config=config)

        self.assertEqual(3, len(account.positions.keys()))

        json = account.json()

        account2 = Account.parse_raw(json)

        self.assertEqual(account.start_date, account2.start_date)

    def test_eval(self):
        config = create_savings_account()

        account_type = config.get_account_type("savingsAccount")
        account = Account(start_date=date(2019, 1, 1), account_type_name=account_type.name,
                          config_version=config.version, config=config)

        self.assertEqual(Decimal(0), account.evaluate("self.current", None))

        self.assertEqual(date(2019, 1, 1), account.evaluate("self.start_date", None))

        self.assertEqual(date(2019, 1, 1) + relativedelta(months=+1) + relativedelta(days=-1),
                         account.evaluate("self.start_date + relativedelta(months=+1) + relativedelta(days=-1)",
                                          None))

    def test_valuation(self):
        config: Configuration = create_savings_account()

        # account = create_account(config, "savingsAccount", date(2019, 1, 1))
        account_type = config.get_account_type("savingsAccount")
        account = Account(start_date=date(2019, 1, 1), account_type_name=account_type.name,
                          config_version= config.version, config=config)

        valuation = AccountValuation(account, account_type, config, date(2020, 1, 1))

        deposit_transaction_type = config.get_transaction_type("deposit")

        external_transactions = group_by_date([
            ExternalTransaction(transaction_type_name=deposit_transaction_type.name,
                                amount= Decimal(1000), value_date=date(2019, 1, 1))])

        valuation.forecast(date(2020, 1, 1), external_transactions)

        self.assertAlmostEqual(account.positions['current'].amount, Decimal(1030.41), places=1)
        self.assertAlmostEqual(account.positions['withholding'].amount, Decimal(30.41) * Decimal(0.2), places=1)


if __name__ == '__main__':
    unittest.main()



from accounts.metadata import *


def create_savings_account() -> Configuration:
    config: Configuration = Configuration(version="v1")

    current = config.add_position_type("current", "current balance")
    interest_accrued = config.add_position_type("accrued", "interest accrued")
    withholding = config.add_position_type("withholding", "withholding tax")

    deposit = config.add_transaction_type("deposit", "Deposit").add_position_rule(TransactionOperation.CREDIT, current)

    interest_accrued_tt = config.add_transaction_type("interestAccrued", "Interest Accrued")
    interest_accrued_tt.add_position_rule(TransactionOperation.CREDIT, interest_accrued)

    capitalized = config.add_transaction_type("capitalized", "Interest Capitalized").add_position_rule(
        TransactionOperation.CREDIT, current).add_position_rule(TransactionOperation.DEBIT, interest_accrued)

    withholding_txn = config.add_transaction_type("withholdingTax", "Withholding Tax").add_position_rule(
        TransactionOperation.CREDIT, withholding)

    savings_account = config.add_account_type("savingsAccount", "Savings Account")

    savings_account.add_transaction_type(deposit)
    savings_account.add_transaction_type(interest_accrued_tt)
    savings_account.add_transaction_type(capitalized)
    savings_account.add_transaction_type(withholding_txn)

    accrual_schedule = ScheduleType(name="accrual", label="Accrual Schedule", frequency=ScheduleFrequency.DAILY,
                                    end_type=ScheduleEndType.NO_END,
                                    business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                    interval_expression="1", start_date_expression="account.start_date")

    savings_account.add_schedule_type(accrual_schedule)

    compounding_schedule = ScheduleType(name="compounding", label="Compounding Schedule", frequency=ScheduleFrequency.MONTHLY,
                                        end_type=ScheduleEndType.NO_END,
                                        business_day_adjustment=BusinessDayAdjustment.NO_ADJUSTMENT,
                                        interval_expression="1",
                                        start_date_expression="account.start_date + relativedelta(month=+1) + relativedelta(days=-1)")

    savings_account.add_schedule_type(compounding_schedule)

    savings_account.add_scheduled_transaction(accrual_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                              interest_accrued_tt,
                                              "account.current * config.interest.get_rate(account.current) / Decimal(365)")

    savings_account.add_scheduled_transaction(compounding_schedule, ScheduledTransactionTiming.END_OF_DAY,
                                              capitalized, "account.accrued")

    interest_rate = config.add_rate_type("interest", "Interest Rate")

    interest_rate.add_tier(Decimal(10000), Decimal(0.03))
    interest_rate.add_tier(Decimal(100000), Decimal(0.035))
    interest_rate.add_tier(Decimal(50000), Decimal(0.04))

    config.add_trigger_transaction(capitalized, withholding_txn, "transaction.amount * Decimal(0.2)")

    return config



import unittest

from accounts.metadata import *
from tests.test_config import create_savings_account


class TestConfiguration(unittest.TestCase):
    def test_serialize(self):
        config = create_savings_account()

        # serialize config to text using Pydantic

        text = config.json()

        config2 = Configuration.parse_raw(text)

        text2 = config2.json()

        self.assertEqual(text, text2)


class RateTest(unittest.TestCase):
    def test_get_max_when_empy(self):
        rate_type = RateType(name='rates', label='Rates')

        max_value = rate_type.get_max_to_amount()

        self.assertEqual(Decimal(0), max_value)

    def test_get_max_when_non_empy(self):
        rate_type = RateType(name='rates', label='Rates')

        rate_type.add_tier(Decimal(100), Decimal(5))

        max_value = rate_type.get_max_to_amount()

        self.assertEqual(Decimal(100), max_value)


if __name__ == '__main__':
    unittest.main()
